#include <stdlib.h>
#include <stdio.h>
#include <string.h>

const double load_factor = 0.5;
const int stride = 1; /* must be odd */

#define DOUBLE_HASH 0

typedef struct str_entry {
  char *str;
  int val;
} str_entry;

typedef struct hash_table {
  int size;
  int count;
  str_entry *a;
} hash_table;

int hash_string(char *s) {
  int i, r = 0;

  for (i = 0; s[i] != 0; i++) {
    r *= 10;
    r += s[i];
  }

  if (r < 0) r = -r;

  return r;
}

int secondary_hash_string(char *s) {
  int i, r = 0;

  for (i = 0; s[i] != 0; i++) {
    r += s[i];
  }

  return r | 1;
}

void resize_hash_table(hash_table *ht);

void reset_hash_table(hash_table *ht, int size)
{
  int i;

  ht->size = size;
  ht->count = 0;

  ht->a = (str_entry*)malloc(ht->size * sizeof(str_entry));
  for (i = 0; i < size; i++)
    ht->a[i].str = NULL;
}

hash_table* make_hash_table()
{
  hash_table *ht;

  ht = (hash_table *)malloc(sizeof(hash_table));
  reset_hash_table(ht, 8);
  
  return ht;
}

int hash_get(hash_table *ht, char *s)
{
  int n = hash_string(s) % ht->size;
#if DOUBLE_HASH
  int n2 = secondary_hash_string(s);
#else
  int n2 = stride;
#endif

  while (1) {
    if (!ht->a[n].str)
      return -1;
    if (!strcmp(ht->a[n].str, s))
      return ht->a[n].val;
    n = (n + n2) % ht->size;
  }
}

void hash_set(hash_table *ht, char *s, int val)
{
  int n = hash_string(s) % ht->size;
#if DOUBLE_HASH
  int n2 = secondary_hash_string(s);
#else
  int n2 = stride;
#endif

  while (1) {
    if (!ht->a[n].str)
      break;
    if (!strcmp(ht->a[n].str, s)) {
      ht->a[n].val = val;
      return;
    }
    n = (n + n2) % ht->size;
  }

  ht->a[n].str = s;
  ht->a[n].val = val;

  ht->count++;

  if (ht->count > ht->size * load_factor)
    resize_hash_table(ht);
}

void resize_hash_table(hash_table *ht) {
  int i, size = ht->size;
  str_entry *a = ht->a;
  
  reset_hash_table(ht, size * 2);
  for (i = 0; i < size; i++) {
    if (a[i].str)
      hash_set(ht, a[i].str, a[i].val);
  }
}

int main(int argc, char** argv)
{
  hash_table* ht;
  char buffer[256];
  int i;
  FILE *f;

  ht = make_hash_table();

  f = fopen(argv[1], "r");
  while (fgets(buffer, 256, f)) {
    hash_set(ht, strdup(buffer), 1);
  }
  fclose(f);

  printf("%d words\n", ht->count);

  while (fgets(buffer, 256, stdin)) {
    i = hash_get(ht, buffer);
    if (i < 0)
      printf("Not there\n");
    else
      printf("Value: %d\n", i);
  }

  return 0;
}
